;;; cl-scopes/forge - may the forge be with you!

;;;; A Forth-like interpreter implemented in Common Lisp.

(defpackage :scopes/forge
  (:use :common-lisp)
  (:export #:*forge-env*
           #:forge-env #:dstack #:exec #:exec-str #:repl
           #:make-seq #:seq-all #:seq-cur #:seq-end #:seq-add))

(in-package :scopes/forge)

(defvar *builtins* (make-hash-table :test 'equalp))

(defclass forge-env ()
  ((data-stack :initform nil
               :reader data-stack
               :accessor data-stack!)
   (vocabulary :initform (list *builtins*)
               :accessor vocabulary)))

(defun forge-env ()
  (let ((fe (make-instance 'forge-env)))
    (push (make-hash-table :test 'equalp) (vocabulary fe))
    fe))

(defvar *forge-env* (forge-env))

(defun dstack()
  (data-stack *forge-env*))

(defun exec-str (s)
  (exec (read-from-string 
             (concatenate 'string "(" s ")"))))

(defun exec (code)
  (dolist (x code)
    (typecase x
      (symbol (funcall (comp1 x)))
      (compiled-function (funcall x))
      (t (pushd x)))))

(defun call (code)
  (dolist (x code)
    (funcall x)))

(defun repl ()
  (do ((input (read-line) (read-line))) ((string= input "q") nil)
    (exec-str input)))

(defun find-word (key)
  (let ((k (string-downcase (symbol-name key))))
    (dolist (voc (vocabulary *forge-env*))
      (let ((v (gethash k voc)))
        (if v (return v))))))

(defun comp (inp)
  (let ((code nil))
    (dolist (item inp)
      (setf code (cons (comp1 item) code)))
    (reverse code)))

(defun comp1 (item)
  (typecase item
    (symbol (find-word item))
    (cons (comp item))
    (t item)))

(defun register (voc key fn)
  (let ((k (if (symbolp key) (symbol-name key) key)))
    (setf (gethash (string-downcase k) voc) fn)))

; built-in primitives

(defun reg-b (key fn) (register *builtins* key fn))

(reg-b "+" #'(lambda () (pushd (+ (popd) (popd)))))
(reg-b "*" #'(lambda () (pushd (* (popd) (popd)))))

(reg-b "dup" #'(lambda () (pushd (car (dstack)))))

(reg-b "?" #'(lambda () (format t "~a~%" (popd))))
(reg-b "??" #'(lambda () (format t "~a~%" (dstack))))

(reg-b "def" #'(lambda () 
  (let* ((name (popd)) 
         (code (comp (popd))))
    (register (voc) name #'(lambda () (call code))))))

(reg-b "const" #'(lambda ()
  (let ((name (popd))
        (value (popd)))
    (register (voc) name #'(lambda () (pushd value))))))

(reg-b "var" #'(lambda ()
  (let ((name (popd))
        (var (list (popd))))
    (register (voc) name #'(lambda () 
      (pushd #'(lambda (fn) 
        (funcall fn var))))))))

(reg-b "get" #'(lambda ()
  (funcall (popd) #'(lambda (x) (pushd (car x))))))

(reg-b "put" #'(lambda ()
  (let ((fn (popd))
        (vl (popd)))
    (funcall fn #'(lambda (x) (setf (car x) vl))))))

; internal definitions

(defun voc () (car (vocabulary *forge-env*)))

(defun popd () (pop (data-stack! *forge-env*)))

(defun pushd (v) (push v (data-stack! *forge-env*)))

; sequence

(defun make-seq ()
  (let* ((start (list nil))
         (p-cur (list start))
         (p-end (list start)))
    #'(lambda (fn) 
        (funcall fn start p-cur p-end))))

(defun seq-all (seq)
  (funcall seq #'(lambda (start cur end) 
                   (declare (ignore cur) (ignore end))
                   start)))

(defun seq-cur (seq)
  (funcall seq #'(lambda (start p-cur end) 
                   (declare (ignore start) (ignore end))
                   (car (car p-cur)))))

(defun seq-end (seq)
  (funcall seq #'(lambda (start cur p-end) 
                   (declare (ignore start) (ignore cur))
                   (car (car p-end)))))

(defun seq-add (seq v)
  (funcall seq #'(lambda (start cur p-end)
                   (declare (ignore cur))
                   (add-to-sequence v start p-end))))

(defun add-to-sequence (v start p-end)
  (if (car start)
    (progn
      (setf (cdr (car p-end)) (list v))
      (pop (car p-end)))
    (setf (car start) v)))

(defun add-to-seq (v p-end)
  (setf (cdr (car p-end)) (list v))
  (pop (car p-end)))
